"""Core utility functions."""

from __future__ import annotations

import sys
import functools
import inspect
import re
from typing import Union, List, Tuple, Any, Literal
from typing_extensions import TypeAlias

from pyarrow.compute import equal,not_equal,greater,less,greater_equal,less_equal,and_,or_
from pyarrow.types import is_struct
from pyarrow import array, scalar, RecordBatch

from ._errors import UnsupportedOperatorError

def _run_before_decorator(before_func):
    """Returns a decorator that runs the specified `before_func` before the wrapped function."""
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            # Extract 'release' argument
            sig = inspect.signature(func)
            bound_args = sig.bind(*args, **kwargs)
            bound_args.apply_defaults()

            if "release" in bound_args.arguments:
                release_value = bound_args.arguments["release"]
                before_func(release_value)

            return func(*args, **kwargs)

        return wrapper
    return decorator

def wrap_functions_with_release(module_name, before_func,func_list):
    """Dynamically wraps all functions with 'release' parameter in the given module."""
    module = sys.modules[module_name]
    
    for name in dir(module):
        attr = getattr(module, name)
        if callable(attr) and name != before_func.__name__ and name in func_list: # Wrap only user-facing functions with "release"
            sig = inspect.signature(attr)
            if "release" in sig.parameters:  
                setattr(module, name, _run_before_decorator(before_func)(attr))

FieldName: TypeAlias = str
OperatorStr: TypeAlias = Literal["==", "!=", "<", "<=", ">", ">=","contains", "isin"]
FilterValue: TypeAlias = Union[str, int, float, List[Any], Tuple[Any, ...], None]
FilterTuple: TypeAlias = Tuple[FieldName, OperatorStr, FilterValue]
FilterGroup: TypeAlias = List[FilterTuple]

FilterStructure: TypeAlias = List[Union[FilterTuple, FilterGroup]] | FilterTuple
"""FilterStructure represents a list of filtering rules for a DataFrame-like object."""

    
def _parse_field_path(field_spec):
    """Parse a field specification that contains nested dictionary access."""
    path_parts = field_spec.split('.')
    base_column = path_parts[0]
    keys = path_parts[1:] if len(path_parts) > 1 else []
    
    return base_column, keys

def _extract_nested_value(array_column, keys):
    """Extract nested values from a dictionary/struct column."""
    current = array_column
    
    for key in keys:
        if is_struct(current.type):
            current = current.field(key)
        else:
            raise ValueError(f"Cannot access field '{key}' in non-struct column")
    
    return current

def _evaluate_condition(batch: RecordBatch, col: str, op:str, val:Any):
    """Return mask for batch based on filter."""
    if '.' in col:
        base_column, keys = _parse_field_path(col)
        # Extract the field value using the path
        if base_column in batch.column_names:
            col_values = _extract_nested_value(batch[base_column], keys)
        else:
            raise ValueError(f"Column '{base_column}' not found in dataset")
    else:
        col_values = batch[col]
    if op == "==":
        return equal(col_values, scalar(val))
    elif op == "!=":
        return not_equal(col_values, scalar(val))
    elif op == ">":
        return greater(col_values, scalar(val))
    elif op == "<":
        return less(col_values, scalar(val))
    elif op == ">=":
        return greater_equal(col_values, scalar(val))
    elif op == "<=":
        return less_equal(col_values, scalar(val))
    elif op == "contains":
        mask = [any(v in value for v in val) if value is not None and isinstance(val, list) else (val in value if value is not None else False) for value in col_values.to_pylist()]
        return array(mask)
    else:
        raise UnsupportedOperatorError(f"Unsupported operator: {op}")
    
def evaluate_filter_structure(batch: RecordBatch, structure):
    """
    Convert a place name to its geometry and bounding box.
    
    Parameters:
    -----------
    batch: PyArrow RecordBatch
        Batch to evaluate.
    structure: FilterStructure
        Filters used to evaluate.
        
        
    Filter structure rules:
    - A tuple (col, op, val) represents a single condition
    - Items at the same level are combined with OR
    - Items nested within a list are combined with AND
    
    Examples:
    - (col, op, val) - Simple condition
    - [tupleA, tupleB] - tupleA OR tupleB
    - [[tupleA, tupleB]] - tupleA AND tupleB  
    - [tupleA, [tupleB, tupleC]] - tupleA OR (tupleB AND tupleC)
    - [[tupleA, tupleB], [tupleC, tupleD]] - (tupleA AND tupleB) OR (tupleC AND tupleD)
        
    Returns:
    --------
    PyArrow Array
        PyArrow boolean array representing the combined filter mask
    """
    # Helper function for type checking
    def is_filter_triplet(item):
        return (isinstance(item, tuple) and len(item) == 3 and 
                isinstance(item[0], str) and isinstance(item[1], str))
    
    # Base case: single filter triplet
    if is_filter_triplet(structure):
        col, op, val = structure
        return _evaluate_condition(batch, col, op, val)
    
    # Must be a list
    if not isinstance(structure, list):
        raise ValueError(f"Invalid filter structure: {structure}")
    
    # All items at same level are OR'd together
    result_masks = []
    
    for item in structure:
        if is_filter_triplet(item):
            # Simple filter condition
            col, op, val = item
            mask = _evaluate_condition(batch, col, op, val)
            result_masks.append(mask)
        elif isinstance(item, list):
            # Nested list - items within are AND'd together
            and_masks = []
            for sub_item in item:
                sub_mask = evaluate_filter_structure(batch, sub_item)
                and_masks.append(sub_mask)
            
            if and_masks:
                # Combine with AND
                combined = and_masks[0]
                for mask in and_masks[1:]:
                    combined = and_(combined,mask)
                result_masks.append(combined)
        else:
            raise ValueError(f"Invalid filter element: {item}")
    
    # Combine all results at this level with OR
    if not result_masks:
        raise ValueError("No valid filters found in structure")
    
    result = result_masks[0]
    for mask in result_masks[1:]:
        result = or_(result,mask)
    
    return batch.filter(result)

def catch_column_filter_error(exc_type: BaseException,error: Exception) -> None:
    """
    Throw user-friendly PyArrow errors.
    
    Parameters:
    -----------
    exc_type : BaseException
        Exception data
    error : Exception
        Exception thrown
    
    Returns:
    --------
    None
    """
    # Capture the full traceback
    error_message = str(error)
    # print(exc_type.__name__)
    if exc_type.__name__ == "UnsupportedOperatorError":
        raise error
    elif exc_type.__name__ == "ArrowInvalid":
        match = re.search(r"FieldRef\.Name\(([^)]+)\)", error_message)
        name = match.group(1)
        raise KeyError(f"Invalid column name:\"{name}\"") from error
    elif exc_type.__name__ =="ArrowNotImplementedError":
        match = re.search(r"\(([^)]+)\)", error_message)
        first_value,last_value = match.group(1).split(",")
        raise ValueError(f"Incorrect type used for value in filter: \"{last_value.strip()}\" should be \"{first_value.strip()}\"") from error


"""Functions to fetch  geoparquet data from Overture Maps on AWS"""
from importlib import resources
from uuid import uuid4
from typing import Union
from geopandas import GeoDataFrame
from pandas import read_csv, DataFrame
from ._utils import FilterStructure, wrap_functions_with_release
from ._io_utils import from_address, from_bbox, from_place, schema_from_dataset
from ._category_finder import CategoryFinder


#TODO latest release reads from text file
OVERTURE_MAIN_PATH = 's3://overturemaps-us-west-2/release/{release}/'
OVERTURE_BUCKET = 'overturemaps-us-west-2'
OVERTURE_REGION = 'us-west-2'
OVERTURE_LATEST_RELEASE = "2025-04-23.0"

OVERTURE_PLACES_PREFIX = "theme=places/type=place/"
OVERTURE_BUILDINGS_PREFIX = "theme=buildings/type=building/"
OVERTURE_BUILDINGS_PART_PREFIX = "theme=buildings/type=building_part/"
OVERTURE_ADDRESSES_PREFIX = "theme=addresses/type=address/"
OVERTURE_TRANSPORTATION_SEGMENT_PREFIX = "theme=transportation/type=segment/"
OVERTURE_TRANSPORTATION_CONNECTOR_PREFIX = "theme=transportation/type=connector/"
OVERTURE_BASE_PREFIX = "theme=base/type={type}"

OVERTURE_CATEGORIES_URL = "https://raw.githubusercontent.com/OvertureMaps/schema/refs/heads/main/docs/schema/concepts/by-theme/places/overture_categories.csv"

def overture_places_from_address(address: str | tuple[float,float],
                                columns: list[str]| None = None,
                                filters: FilterStructure | None = None,
                                distance: float = 500 ,
                                unit: str = "m" ,
                                release: str = OVERTURE_LATEST_RELEASE) -> GeoDataFrame:
    """
    Retrieve places data from Overture in a bounding box around a specified address.
    
    Parameters
    ----------
    address : str | tuple[float,float]
        The address or (longitude, latitude) tuple to search for nearby places.
    columns : list[str] | None, optional
        Specific columns to include in the result, by default None
    filters : FilterStructure | None, optional
        Filter criteria to apply to the results. By default, None.
        Should be a list in the format(column,operator,value)
        Supported operators are: "==", "!=", "<", "<=", ">", ">=","contains"
    distance : float, optional
        Radius of the bounding box around the address. Defaults to 500 meters.
    unit : str, optional
        Unit of measurement for the distance. Defaults to "m" (meters). One of: "m","km","in","ft","yd","mi"
    release : str, optional
        Dataset release version to use. Defaults to the latest version.
        
    Returns
    -------
    GeoDataFrame
        A GeoDataFrame containing places data within the specified bounding box of the address
    """
    return from_address(address,OVERTURE_PLACES_PREFIX,OVERTURE_MAIN_PATH,OVERTURE_REGION,release,columns,filters,distance,unit)

def overture_places_from_place(address: str,columns: list[str]| None=None,filters: FilterStructure| None=None,release: str=OVERTURE_LATEST_RELEASE)-> GeoDataFrame:
    """
    Retrieve places data from Overture by its address or place name.
    
    Parameters
    ----------
    address : str
        The place name to search within
    columns : list[str] | None, optional
        Specific columns to include in the result, by default None
    filters : FilterStructure | None, optional
        Filter criteria to apply to the results. By default, None.
        Should be a list in the format(column,operator,value)
        Supported operators are: "==", "!=", "<", "<=", ">", ">=","contains"
    release : str, optional
        Dataset release version to use. Defaults to the latest version.
        
    Returns
    -------
    GeoDataFrame
        A GeoDataFrame containing places data within the specified place
    """
    return from_place(address,OVERTURE_PLACES_PREFIX,OVERTURE_MAIN_PATH,OVERTURE_REGION,release,columns,filters)

def overture_places_from_bbox(bbox: tuple[float,float,float,float],columns: list[str]| None=None,filters: FilterStructure| None=None,release: str=OVERTURE_LATEST_RELEASE)-> GeoDataFrame:
    """
    Retrieve places data from Overture within a bounding box.
    
    Parameters
    ----------
    bbox : tuple[float, float, float, float]
        The bounding box coordinates (min_x, min_y, max_x, max_y)
    columns : list[str] | None, optional
        Specific columns to include in the result, by default None
    filters : FilterStructure | None, optional
        Filter criteria to apply to the results. By default, None.
        Should be a list in the format(column,operator,value)
        Supported operators are: "==", "!=", "<", "<=", ">", ">=","contains"
    release : str, optional
        Dataset release version to use. Defaults to the latest version.
        
    Returns
    -------
    GeoDataFrame
        A GeoDataFrame containing places data within the specified bounding box
    """
    return from_bbox(bbox,OVERTURE_PLACES_PREFIX,OVERTURE_MAIN_PATH,OVERTURE_REGION,release,columns,filters)

def overture_buildings_from_address(address: str | tuple[float,float],columns: list[str]| None = None,filters: FilterStructure| None = None,distance: float = 500 ,unit: str = "m" ,release: str = OVERTURE_LATEST_RELEASE,building_part: bool=False) -> GeoDataFrame:
    """
    Retrieve buildings data from Overture in a bounding box around a specified address.
    
    Parameters
    ----------
    address : str | tuple[float,float]
        The address or (longitude, latitude) tuple to search for nearby places.
    columns : list[str] | None, optional
        Specific columns to include in the result, by default None
    filters : FilterStructure | None, optional
        Filter criteria to apply to the results. By default, None.
        Should be a list in the format(column,operator,value)
        Supported operators are: "==", "!=", "<", "<=", ">", ">=","contains"
    distance : float, optional
        Radius of the bounding box around the address. Defaults to 500 meters.
    unit : str, optional
        Unit of measurement for the distance. Defaults to "m" (meters). One of: "m","km","in","ft","yd","mi"
    release : str, optional
        Dataset release version to use. Defaults to the latest version.
    building_part : bool, optional
        Whether to retrieve building parts instead of whole buildings, by default False
        
    Returns
    -------
    GeoDataFrame
        A GeoDataFrame containing buildings data within the specified bounding box of the address
    """
    if building_part:
        prefix = OVERTURE_BUILDINGS_PART_PREFIX
    else:
        prefix = OVERTURE_BUILDINGS_PREFIX 
    return from_address(address,prefix,OVERTURE_MAIN_PATH,OVERTURE_REGION,release,columns,filters,distance,unit)

def overture_buildings_from_place(address: str,columns: list[str]| None=None,filters: FilterStructure| None=None,release: str=OVERTURE_LATEST_RELEASE,building_part: bool=False)-> GeoDataFrame:
    """
    Retrieve buildings data from Overture by its address or place name.
    
    Parameters
    ----------
    address : str
        The place name to search within
    columns : list[str] | None, optional
        Specific columns to include in the result, by default None
    filters : FilterStructure | None, optional
        Filter criteria to apply to the results. By default, None.
        Should be a list in the format(column,operator,value)
        Supported operators are: "==", "!=", "<", "<=", ">", ">=","contains"
    release : str, optional
        Dataset release version to use. Defaults to the latest version.
    building_part : bool, optional
        Whether to retrieve building parts instead of whole buildings, by default False
        
    Returns
    -------
    GeoDataFrame
        A GeoDataFrame containing buildings data within the specified place
    """
    if building_part:
        prefix = OVERTURE_BUILDINGS_PART_PREFIX
    else:
        prefix = OVERTURE_BUILDINGS_PREFIX
    return from_place(address,prefix,OVERTURE_MAIN_PATH,OVERTURE_REGION,release,columns,filters)

def overture_buildings_from_bbox(bbox: tuple[float,float,float,float],columns: list[str]| None=None,filters: FilterStructure| None=None,release: str=OVERTURE_LATEST_RELEASE,building_part: bool=False)-> GeoDataFrame:
    """
    Retrieve buildings data from Overture within a bounding box.
    
    Parameters
    ----------
    bbox : tuple[float, float, float, float]
        The bounding box coordinates (min_x, min_y, max_x, max_y)
    columns : list[str] | None, optional
        Specific columns to include in the result, by default None
    filters : FilterStructure | None, optional
        Filter criteria to apply to the results. By default, None.
        Should be a list in the format(column,operator,value)
        Supported operators are: "==", "!=", "<", "<=", ">", ">=","contains"
    release : str, optional
        Dataset release version to use. Defaults to the latest version.
    building_part : bool, optional
        Whether to retrieve building parts instead of whole buildings, by default False
        
    Returns
    -------
    GeoDataFrame
        A GeoDataFrame containing buildings data within the specified bounding box
    """
    if building_part:
        prefix = OVERTURE_BUILDINGS_PART_PREFIX
    else:
        prefix = OVERTURE_BUILDINGS_PREFIX
    return from_bbox(bbox,prefix,OVERTURE_MAIN_PATH,OVERTURE_REGION,release,columns,filters)

def overture_transportation_from_address(address: str | tuple[float,float],columns: list[str]| None = None,filters: FilterStructure| None = None,distance: float = 500 ,unit: str = "m" ,release: str = OVERTURE_LATEST_RELEASE,connector:bool=False) -> GeoDataFrame:
    """
    Retrieve transportation data from Overture in a bounding box around a specified address.
    
    Parameters
    ----------
    address : str | tuple[float,float]
        The address or (longitude, latitude) tuple to search for nearby places.
    columns : list[str] | None, optional
        Specific columns to include in the result, by default None
    filters : FilterStructure | None, optional
        Filter criteria to apply to the results. By default, None.
        Should be a list in the format(column,operator,value)
        Supported operators are: "==", "!=", "<", "<=", ">", ">=","contains"
    distance : float, optional
        Radius of the bounding box around the address. Defaults to 500 meters.
    unit : str, optional
        Unit of measurement for the distance. Defaults to "m" (meters). One of: "m","km","in","ft","yd","mi"
    release : str, optional
        Dataset release version to use. Defaults to the latest version.
    connector : bool, optional
        Whether to retrieve connectors instead of segments, by default False
        
    Returns
    -------
    GeoDataFrame
        A GeoDataFrame containing transportation data within the specified bounding box of the address
    """
    if connector:
        prefix = OVERTURE_TRANSPORTATION_CONNECTOR_PREFIX
    else:
        prefix = OVERTURE_TRANSPORTATION_SEGMENT_PREFIX
    return from_address(address,prefix,OVERTURE_MAIN_PATH,OVERTURE_REGION,release,columns,filters,distance,unit)

def overture_transportation_from_place(address: str,columns: list[str]| None=None,filters: FilterStructure| None=None,release: str=OVERTURE_LATEST_RELEASE,connector:bool=False)-> GeoDataFrame:
    """
    Retrieve transportation data from Overture by its address or place name.
    
    Parameters
    ----------
    address : str
        The place name to search within
    columns : list[str] | None, optional
        Specific columns to include in the result, by default None
    filters : FilterStructure | None, optional
        Filter criteria to apply to the results. By default, None.
        Should be a list in the format(column,operator,value)
        Supported operators are: "==", "!=", "<", "<=", ">", ">=","contains"
    release : str, optional
        Dataset release version to use. Defaults to the latest version.
    connector : bool, optional
        Whether to retrieve connectors instead of segments, by default False
        
    Returns
    -------
    GeoDataFrame
        A GeoDataFrame containing transportation data within the specified place
    """
    if connector:
        prefix = OVERTURE_TRANSPORTATION_CONNECTOR_PREFIX
    else:
        prefix = OVERTURE_TRANSPORTATION_SEGMENT_PREFIX
    return from_place(address,prefix,OVERTURE_MAIN_PATH,OVERTURE_REGION,release,columns,filters)

def overture_transportation_from_bbox(bbox: tuple[float,float,float,float],columns: list[str]| None=None,filters: FilterStructure | None=None,release: str=OVERTURE_LATEST_RELEASE,connector:bool=False)-> GeoDataFrame:
    """
    Retrieve transportation data from Overture within a bounding box.
    
    Parameters
    ----------
    bbox : tuple[float, float, float, float]
        The bounding box coordinates (min_x, min_y, max_x, max_y)
    columns : list[str] | None, optional
        Specific columns to include in the result, by default None
    filters : FilterStructure | None, optional
        Filter criteria to apply to the results. By default, None.
        Should be a list in the format(column,operator,value)
        Supported operators are: "==", "!=", "<", "<=", ">", ">=","contains"
    release : str, optional
        Dataset release version to use. Defaults to the latest version.
    connector : bool, optional
        Whether to retrieve connectors instead of segments, by default False
        
    Returns
    -------
    GeoDataFrame
        A GeoDataFrame containing transportation data within the specified bounding box
    """
    if connector:
        prefix = OVERTURE_TRANSPORTATION_CONNECTOR_PREFIX
    else:
        prefix = OVERTURE_TRANSPORTATION_SEGMENT_PREFIX
    return from_bbox(bbox,prefix,OVERTURE_MAIN_PATH,OVERTURE_REGION,release,columns,filters)
    
def overture_addresses_from_address(address: str | tuple[float,float],columns: list[str]| None = None,filters: FilterStructure| None = None,distance: float = 500 ,unit: str = "m" ,release: str = OVERTURE_LATEST_RELEASE) -> GeoDataFrame:
    """
    Retrieve address data from Overture in a bounding box around a specified address.
    
    Parameters
    ----------
    address : str | tuple[float,float]
        The address or (longitude, latitude) tuple to search for nearby places.
    columns : list[str] | None, optional
        Specific columns to include in the result, by default None
    filters : FilterStructure | None, optional
        Filter criteria to apply to the results. By default, None.
        Should be a list in the format(column,operator,value)
        Supported operators are: "==", "!=", "<", "<=", ">", ">=","contains"
    distance : float, optional
        Radius of the bounding box around the address. Defaults to 500 meters.
    unit : str, optional
        Unit of measurement for the distance. Defaults to "m" (meters). One of: "m","km","in","ft","yd","mi"
    release : str, optional
        Dataset release version to use. Defaults to the latest version.
        
    Returns
    -------
    GeoDataFrame
        A GeoDataFrame containing address data within the specified bounding box of the address
    """
    return from_address(address,OVERTURE_ADDRESSES_PREFIX,OVERTURE_MAIN_PATH,OVERTURE_REGION,release,columns,filters,distance,unit)

def overture_addresses_from_place(address: str,columns: list[str]| None=None,filters: FilterStructure| None=None,release: str=OVERTURE_LATEST_RELEASE)-> GeoDataFrame:
    """
    Retrieve address data from Overture by its address or place name.
    
    Parameters
    ----------
    address : str
        The place name to search within
    columns : list[str] | None, optional
        Specific columns to include in the result, by default None
    filters : FilterStructure | None, optional
        Filter criteria to apply to the results. By default, None.
        Should be a list in the format(column,operator,value)
        Supported operators are: "==", "!=", "<", "<=", ">", ">=","contains"
    release : str, optional
        Dataset release version to use. Defaults to the latest version.
        
    Returns
    -------
    GeoDataFrame
        A GeoDataFrame containing address data within the specified place
    """
    return from_place(address,OVERTURE_ADDRESSES_PREFIX,OVERTURE_MAIN_PATH,OVERTURE_REGION,release,columns,filters)

def overture_addresses_from_bbox(bbox: tuple[float,float,float,float],columns: list[str]| None=None,filters: FilterStructure| None=None,release: str=OVERTURE_LATEST_RELEASE)-> GeoDataFrame:
    """
    Retrieve address data from Overture within a bounding box.
    
    Parameters
    ----------
    bbox : tuple[float, float, float, float]
        The bounding box coordinates (min_x, min_y, max_x, max_y)
    columns : list[str] | None, optional
        Specific columns to include in the result, by default None
    filters : FilterStructure | None, optional
        Filter criteria to apply to the results. By default, None.
        Should be a list in the format(column,operator,value)
        Supported operators are: "==", "!=", "<", "<=", ">", ">=","contains"
    release : str, optional
        Dataset release version to use. Defaults to the latest version.
        
    Returns
    -------
    GeoDataFrame
        A GeoDataFrame containing address data within the specified bounding box
    """
    return from_bbox(bbox,OVERTURE_ADDRESSES_PREFIX,OVERTURE_MAIN_PATH,OVERTURE_REGION,release,columns,filters)

def overture_base_from_address(address: str | tuple[float,float],base_type: str,columns: list[str]| None = None,filters: FilterStructure| None = None,distance: float = 500 ,unit: str = "m" ,release: str = OVERTURE_LATEST_RELEASE) -> GeoDataFrame:
    """
    Retrieve base data of a specific type from Overture in a bounding box around a specified address.
    
    Parameters
    ----------
    address : str | tuple[float,float]
        The address or (longitude, latitude) tuple to search for nearby places.
    base_type : str
        The type of base data to retrieve. One of: "bathymetry","infrastructure","land","land_cover",land_use","water".
    columns : list[str] | None, optional
        Specific columns to include in the result, by default None
    filters : FilterStructure | None, optional
        Filter criteria to apply to the results. By default, None.
        Should be a list in the format(column,operator,value)
        Supported operators are: "==", "!=", "<", "<=", ">", ">=","contains"
    distance : float, optional
        Radius of the bounding box around the address. Defaults to 500 meters.
    unit : str, optional
        Unit of measurement for the distance. Defaults to "m" (meters). One of: "m","km","in","ft","yd","mi"
    release : str, optional
        Dataset release version to use. Defaults to the latest version.
        
    Returns
    -------
    GeoDataFrame
        A GeoDataFrame containing base data of the specified type within the specified bounding box of the address.
    """
    _check_base_type(base_type)
    complete_prefix = OVERTURE_BASE_PREFIX.format(type=base_type)
    return from_address(address,complete_prefix,OVERTURE_MAIN_PATH,OVERTURE_REGION,release,columns,filters,distance,unit)
    

def overture_base_from_place(address: str,base_type: str,columns: list[str]| None=None,filters: FilterStructure | None =None,release: str=OVERTURE_LATEST_RELEASE)-> GeoDataFrame:
    """
    Retrieve base data of a specific type from Overture by its address or place name.
    
    Parameters
    ----------
    address : str
        The place name to search within
    base_type : str
        The type of base data to retrieve. One of: "bathymetry","infrastructure","land","land_cover",land_use","water"
    columns : list[str] | None, optional
        Specific columns to include in the result, by default None
    filters : FilterStructure | None, optional
        Filter criteria to apply to the results. By default, None.
        Should be a list in the format(column,operator,value)
        Supported operators are: "==", "!=", "<", "<=", ">", ">=","contains"
    release : str, optional
        Dataset release version to use. Defaults to the latest version.
        
    Returns
    -------
    GeoDataFrame
        A GeoDataFrame containing base data of the specified type within the place
    """
    _check_base_type(base_type)
    complete_prefix = OVERTURE_BASE_PREFIX.format(type=base_type)
    return from_place(address,complete_prefix,OVERTURE_MAIN_PATH,OVERTURE_REGION,release,columns,filters)
    

def overture_base_from_bbox(bbox: tuple[float,float,float,float],base_type: str,columns: list[str]| None=None,filters: FilterStructure| None=None,release: str=OVERTURE_LATEST_RELEASE)-> GeoDataFrame  :
    """
    Retrieve base data of a specific type from Overture within a bounding box.
    
    Parameters
    ----------
    bbox : tuple[float, float, float, float]
        The bounding box coordinates (min_x, min_y, max_x, max_y)
    base_type : str
        The type of base data to retrieve. One of: "bathymetry","infrastructure","land","land_cover",land_use","water"
    columns : list[str] | None, optional
        Specific columns to include in the result, by default None
    filters : FilterStructure | None, optional
        Filter criteria to apply to the results. By default, None.
        Should be a list in the format(column,operator,value)
        Supported operators are: "==", "!=", "<", "<=", ">", ">=","contains"
    release : str, optional
        Dataset release version to use. Defaults to the latest version.
        
    Returns
    -------
    GeoDataFrame
        A GeoDataFrame containing base data of the specified type within the bounding box
    """
    _check_base_type(base_type)
    complete_prefix = OVERTURE_BASE_PREFIX.format(type=base_type)
    return from_bbox(bbox,complete_prefix,OVERTURE_MAIN_PATH,OVERTURE_REGION,release,columns,filters)

def get_categories():
    
    df = read_csv(OVERTURE_CATEGORIES_URL, converters={" Overture Taxonomy": lambda x : x.strip().strip("[]").split(",")}, sep=';')
    
    # Create a dictionary to store unique IDs for each category
    category_ids = {}
    
    all_records = []
    
    for _, row in df.iterrows():
        category_hierarchy = row[" Overture Taxonomy"]
        
        # Only create a record for the last/deepest level in the hierarchy
        deepest_level = len(category_hierarchy)
        current_category = category_hierarchy[deepest_level - 1]
        
        for category in category_hierarchy:
            if category not in category_ids:
                u_id = str(uuid4()).replace('-', '')[:24]
                while u_id in category_ids.values():
                    u_id = str(uuid4()).replace('-', '')[:24]
                category_ids[category] = u_id
        
        # Build the full path string
        path_parts = []
        for category in category_hierarchy:
            formatted_part = ' '.join(word.capitalize() for word in category.split('_'))
            path_parts.append(formatted_part)
        
        full_path = ' > '.join(path_parts)
        
        record = {
            'category_id': category_ids[current_category],
            'category_level': deepest_level,
            'category_name': current_category,
            'category_label': full_path,
            'level1_category_id': None,
            'level1_category_name': None,
            'level2_category_id': None,
            'level2_category_name': None,
            'level3_category_id': None,
            'level3_category_name': None,
            'level4_category_id': None,
            'level4_category_name': None,
            'level5_category_id': None,
            'level5_category_name': None,
            'level6_category_id': None,
            'level6_category_name': None
        }
        
        # Fill in the level fields based on the hierarchy
        for j in range(min(deepest_level, 6)):
            level_name = category_hierarchy[j]
            level_id = category_ids[level_name]
            
            record[f'level{j+1}_category_id'] = level_id
            record[f'level{j+1}_category_name'] = level_name
        
        all_records.append(record)
    
    result_df = DataFrame(all_records)
    
    column_order = [
        'category_id', 'category_level', 'category_name', 'category_label',
        'level1_category_id', 'level1_category_name',
        'level2_category_id', 'level2_category_name',
        'level3_category_id', 'level3_category_name',
        'level4_category_id', 'level4_category_name',
        'level5_category_id', 'level5_category_name',
        'level6_category_id', 'level6_category_name'
    ]
    
    
    available_columns = [col for col in column_order if col in result_df.columns]
    
    result_df = result_df[available_columns]
    
    return result_df

def get_schema(dataset_name : str,
                connector : bool = False,
                building_part : bool = False,base_type : str = None,release: str = OVERTURE_LATEST_RELEASE):
    """
    Get Arrow schema for the given dataset.

    Parameters
    ----------
        dataset_name : str
            Name of the dataset to get the schema of, must be one of \"buildings\",\"transportation\",\"base\",\"places\",\"addresses\"
        connector : bool, optional
            Whether to retrieve connector schema, by default False.
        building_part : bool, optional 
            Whether to retrieve building_part schema, by default False.
        base_type : str, optional 
            Which base type schema to retrive, by default None.
        release : str, optional 
            Release version to use, defaults to the latest version.
    Returns
    -------
    str
        String representation of PyArrow schema of dataset
    """
    datasets = ["buildings","transportation","base","places","addresses"]
    with resources.files("pyplaces").joinpath("releases/overture/base_types.txt").open( "r",encoding="utf-8-sig") as f:
        base_types = [line.replace("type=", "").strip(" \n/") for line in f]
    if dataset_name not in datasets:
        raise KeyError(f"No dataset: {dataset_name} found")
    elif connector and dataset_name != "transportation":
        raise KeyError("Dataset must be \"transportation\" to get connector schema")
    elif building_part and dataset_name != "buildings":
        raise KeyError("Dataset must be \"buildings\" to get building_part schema")
    elif base_type and dataset_name != "base":
        raise KeyError(f"Dataset must be \"base\" to get {base_type} schema")
    elif base_type and base_type not in base_types:
        raise KeyError(f"No base type:{dataset_name} found")
    
    path = OVERTURE_MAIN_PATH.format(release=release).replace("s3://", "") 
    if dataset_name == "places":
        path = path + OVERTURE_PLACES_PREFIX
    elif dataset_name == "addresses":
        path = path + OVERTURE_ADDRESSES_PREFIX
    elif dataset_name == "buildings":
        if building_part:
            path = path + OVERTURE_BUILDINGS_PART_PREFIX
        else:
            path = path + OVERTURE_BUILDINGS_PREFIX
    elif dataset_name == "base":
        path = path + OVERTURE_BASE_PREFIX.format(type=base_type)
    elif dataset_name == "transportation":
        if connector:
            path = path + OVERTURE_TRANSPORTATION_CONNECTOR_PREFIX
        else:
            path = path + OVERTURE_TRANSPORTATION_SEGMENT_PREFIX
    # print(path)
    schema = schema_from_dataset(path,OVERTURE_REGION)
    return schema.to_string()

def find_categories(search: str, num_results: int = 5, exact_match: bool=False,verbose: bool=False,as_df: bool= False) -> Union[list[str],DataFrame]:
    finder = CategoryFinder()
    categories = get_categories()
    finder.load_data(categories)
    finder.process_data()
    matches = finder.suggest_categories(search,num_results,exact_match,verbose,as_df)
    return matches

#TODO no automation
def _check_base_type(base_type):
    """
    Validate if the provided base type is valid.
    
    Parameters
    ----------
    base_type : str
        The base type to validate
        
    Raises
    ------
    ValueError
        If the base type is not valid
    """   
    with resources.files("pyplaces").joinpath("releases/overture/base_types.txt").open( "r",encoding="utf-8-sig") as f:
        base_types = [line.replace("type=", "").strip(" \n/") for line in f]
    if base_type not in base_types:
        raise ValueError(f"Invalid base type:{base_type}")
    
def _check_release(release):
    """
    Validate if the provided Overture release is valid.
    
    Parameters
    ----------
    release : str
        The release version to validate
        
    Raises
    ------
    ValueError
        If the release version is not valid
    """
    with resources.files("pyplaces").joinpath("releases/overture/releases.txt").open( "r",encoding="utf-8-sig") as f:
        folders = [line.strip(" \n/") for line in f]
    if release not in folders:
        raise ValueError(f"Invalid release:{release}")
    


__all__ = ["overture_addresses_from_address","overture_addresses_from_bbox","overture_addresses_from_place","overture_base_from_address", 
            "overture_base_from_bbox","overture_base_from_place","overture_buildings_from_address","overture_buildings_from_bbox",
            "overture_buildings_from_place","overture_places_from_address","overture_places_from_bbox","overture_places_from_place",
            "overture_transportation_from_address","overture_transportation_from_bbox","overture_transportation_from_place","get_schema","find_categories"]

wrap_functions_with_release(__name__, _check_release,__all__)



"""Functions to fetch  geoparquet data from Foursquare Open Places on AWS"""
from importlib import resources
from typing import Union
from geopandas import GeoDataFrame
from pandas import DataFrame
from ._utils import FilterStructure, wrap_functions_with_release
from ._io_utils import from_address, from_bbox, from_place, read_parquet_arrow, schema_from_dataset
from ._category_finder import CategoryFinder


#TODO latest release reads from text file
FSQ_MAIN_PATH = 's3://fsq-os-places-us-east-1/release/dt={release}/'
FSQ_BUCKET = 'fsq-os-places-us-east-1'
FSQ_REGION = 'us-east-1'
FSQ_PLACES_PREFIX = "places/parquet/"
FSQ_CATEGORIES_PREFIX = "categories/parquet/"
FSQ_LATEST_RELEASE = "2025-04-08"

# FSQ_FUSED_MAIN_PATH = 's3://us-west-2.opendata.source.coop/fused/fsq-os-places/{release}/'
# FSQ_FUSED_BUCKET = 's-west-2.opendata.source.coop'
# FSQ_FUSED_REGION = 'us-west-2'
# FSQ_FUSED_LATEST_RELEASE = "2025-04-08"
# FSQ_FUSED_PLACES_PREFIX = "places/"

def foursquare_places_from_address(address: str | tuple[float,float],
                                    columns: list[str] | None = None,
                                    filters: FilterStructure | None = None,
                                    distance: float = 500,
                                    unit: str = "m",
                                    release: str = FSQ_LATEST_RELEASE) -> GeoDataFrame:
    """
    Retrieves Foursquare places data in a bounding box around a specified address.
    
    Parameters
    ----------
    address : str | tuple[float,float]
        The addres or (longitude, latitude) tuple to search for nearby places.
    columns : list[str] | None, optional
        Specific columns to retrieve from the dataset. If None, all columns are retrieved.
    filters : FilterStructure | None, optional
        Filter criteria to apply to the results. By default, None.
        Should be a list in the format(column,operator,value)
        Supported operators are: "==", "!=", "<", "<=", ">", ">=","contains"
    distance : float, default 500
        Radius of the bounding box around the address. Defaults to 500 meters.
    unit : str, default "m"
        Unit of measurement for the distance. Defaults to "m" (meters). One of: "m","km","in","ft","yd","mi"
    release : str, default FSQ_LATEST_RELEASE
        Dataset release version to use. Defaults to the latest version.
        
    Returns
    -------
    GeoDataFrame
        A GeoDataFrame containing Foursquare places data within the specified bounding box of the address.
    """
    return from_address(address, FSQ_PLACES_PREFIX, FSQ_MAIN_PATH, FSQ_REGION, release, columns, filters, distance, unit)


def foursquare_places_from_place(address: str,
                                    columns: list[str] | None = None, 
                                    filters: FilterStructure = None, 
                                    release: str = FSQ_LATEST_RELEASE) -> GeoDataFrame:
    """
    Retrieves Foursquare places data for a specific place identified by its address or place name.
    
    Parameters
    ----------
    address : str
        The address or identifier of the place to retrieve data for.
    columns : list[str] | None, optional
        Specific columns to retrieve from the dataset. If None, all columns are retrieved.
    filters : FilterStructure | None, optional
        Filter criteria to apply to the results. By default, None.
        Should be a list in the format(column,operator,value)
        Supported operators are: "==", "!=", "<", "<=", ">", ">=","contains"
    release : str, default FSQ_LATEST_RELEASE
        Dataset release version to use. Defaults to the latest version.
        
    Returns
    -------
    GeoDataFrame
        A GeoDataFrame containing Foursquare places data within specified place.
    """
    return from_place(address, FSQ_PLACES_PREFIX, FSQ_MAIN_PATH, FSQ_REGION, release, columns, filters)


def foursquare_places_from_bbox(bbox: tuple[float, float, float, float], 
                                columns: list[str] | None = None, 
                                filters: FilterStructure | None = None, 
                                release: str = FSQ_LATEST_RELEASE) -> GeoDataFrame:
    """
    Retrieves Foursquare places data within a specified bounding box.
    
    Parameters
    ----------
    bbox : tuple[float, float, float, float]
        Bounding box coordinates in the format (min_x, min_y, max_x, max_y).
    columns : list[str] | None, optional
        Specific columns to retrieve from the dataset. If None, all columns are retrieved.
    filters : FilterStructure | None, optional
        Filter criteria to apply to the results. By default, None.
        Should be a list in the format(column,operator,value)
        Supported operators are: "==", "!=", "<", "<=", ">", ">=","contains"
    release : str, default FSQ_LATEST_RELEASE
        Dataset release version to use. Defaults to the latest version.
        
    Returns
    -------
    GeoDataFrame
        A GeoDataFrame containing Foursquare places data within the specified bounding box.
    """
    return from_bbox(bbox, FSQ_PLACES_PREFIX, FSQ_MAIN_PATH, FSQ_REGION, release, columns, filters)


def get_categories(columns: list[str] | None = None,
                    filters: FilterStructure | None = None,
                    release: str = FSQ_LATEST_RELEASE) -> DataFrame:
    """
    Retrieves Foursquare place categories data.
    
    Parameters
    ----------
    columns : list[str] | None, optional
        Specific columns to retrieve from the dataset. If None, all columns are retrieved.
    filters : FilterStructure | None, optional
        Filter criteria to apply to the results. By default, None.
        Should be a list in the format(column,operator,value)
        Supported operators are: "==", "!=", "<", "<=", ">", ">=","contains"
    release : str, default FSQ_LATEST_RELEASE
        Dataset release version to use. Defaults to the latest version.
        
    Returns
    -------
    DataFrame
        A DataFrame containing Foursquare place categories data.
        
    Notes
    -----
    This function constructs the data path using Foursquare-specific constants and
    retrieves the categories data using read_parquet_arrow.
    """
    path = FSQ_MAIN_PATH.format(release=release) + FSQ_CATEGORIES_PREFIX
    return read_parquet_arrow(path, FSQ_REGION, columns, filters)

def find_categories(search: str, num_results: int = 5, exact_match: bool=False,verbose: bool=False,as_df: bool= False) -> Union[list[str],DataFrame]:
    finder = CategoryFinder()
    categories = get_categories()
    finder.load_data(categories)
    finder.process_data()
    matches = finder.suggest_categories(search,num_results,exact_match,verbose,as_df)
    return matches

def _check_release(release: str):
    """
    Validates if the specified Foursquare data release version exists.
    
    Parameters
    ----------
    release : str
        The release version to validate.
        
    Raises
    ------
    ValueError
        If the specified release version does not exist in the available releases.
        
    Notes
    -----
    This private helper function reads available release versions from a text file
    and checks if the specified release is among them.
    """
    with resources.files("pyplaces").joinpath("releases/foursquare/releases.txt").open("r", encoding="utf-8-sig") as f:
        folders = [line.replace("dt=", "").strip(" \n/") for line in f]
    if release not in folders:
        raise ValueError(f"Invalid release: {release}")
    
def get_schema(categories=False,release:str=FSQ_LATEST_RELEASE) -> str:
    """
    Get Arrow schema for the given dataset. Set categories to True if you want to get the categories instead of the places schema.

    Parameters
    ----------
        connector : bool, optional
            Whether to get categories schema, by default False.
        release : str, optional 
            Release version to use, defaults to the latest version.
    Returns
    -------
    str
        String representation of PyArrow schema of dataset
    """
    path = FSQ_MAIN_PATH.format(release=release).replace("s3://", "") 
    if categories:
        path = path + FSQ_CATEGORIES_PREFIX
    else:
        path = path + FSQ_PLACES_PREFIX
    schema = schema_from_dataset(path,FSQ_REGION)
    return schema.to_string()


__all__ = ["foursquare_places_from_address", "foursquare_places_from_bbox", "foursquare_places_from_place", "get_categories","get_schema","find_categories"]

wrap_functions_with_release(__name__, _check_release,__all__)


"""I/O utility functions for reading data."""

from __future__ import annotations
from json import loads
import sys

from pyarrow import RecordBatchReader
from pyarrow.compute import field
from pyarrow.dataset import dataset
from pyarrow.fs import S3FileSystem
from geopandas import GeoDataFrame

from ._utils import evaluate_filter_structure, catch_column_filter_error, FilterStructure
from ._geo_utils import geocode_place_to_bbox, geocode_point_to_bbox
from ._errors import S3ReadError

def schema_from_dataset(s3_path,region):
    """
    Get schema from PyArrow dataset.
    
    Parameters:
    -----------
    s3_path : str 
        S3 path to directory containing parquet dataset
    region : str
        AWS region
    Returns:
    --------
    Schema
        PyArrow schema from given dataset.
    """
    ds = dataset(
            s3_path, filesystem=S3FileSystem(anonymous=True, region=region)
        )
    return ds.schema

def _decode_bytes(obj):
    if isinstance(obj, dict):
        return {_decode_bytes(k): _decode_bytes(v) for k, v in obj.items()}
    elif isinstance(obj, list):
        return [_decode_bytes(i) for i in obj]
    elif isinstance(obj, bytes):
        return obj.decode("utf-8")
    else:
        return obj

def read_geoparquet_arrow(path: str, region: str, bbox: tuple[float,float,float,float], 
                        columns: list[str] | None = None, 
                        filters: FilterStructure | None = None) -> GeoDataFrame:
    """
    Read geospatial data from a parquet file on S3 with filtering by bbox.
    
    Parameters:
    -----------
    path : str
        S3 path to the parquet file
    region : str
        AWS region
    bbox : tuple
        Bounding box as (minx, miny, maxx, maxy)
    columns : list, optional
        Columns to select
    filters : FilterStructure, optional
        Filter expression
        
    Returns:
    --------
    GeoDataFrame
        Filtered geodataframe
    """
    # get pyarrow dataset
    clean_path = path.replace("s3://", "")
    try:
        ds = dataset(
            clean_path, filesystem=S3FileSystem(anonymous=True, region=region)
        )
    except Exception as e:
        raise S3ReadError(f"Read from bucket {clean_path} could not be complete.") from e
    
    xmin, ymin, xmax, ymax = bbox
    
    geo_filter_expr = (
        (field("bbox", "xmin") < xmax)
        & (field("bbox", "xmax") > xmin)
        & (field("bbox", "ymin") < ymax)
        & (field("bbox", "ymax") > ymin)
    )
    
    
    # filter with bounding box
    try:
        batches = ds.to_batches(filter=geo_filter_expr)
    except Exception as e:
        exc_info = sys.exc_info()[0]
        catch_column_filter_error(exc_info,e)
        
    non_empty_batches = (b for b in batches if b.num_rows > 0)
    
    
    
    def filter_batches(batches):
        for b in batches:
            yield evaluate_filter_structure(b,filters)
    
    # generate results from complex filters(if needed)
    if filters:
        filtered_batches = filter_batches(non_empty_batches)
    else:
        filtered_batches = non_empty_batches
    # convert geometry column to correct type in schema
    schema = ds.schema
    metadata_str = _decode_bytes(schema.metadata) 
    geo_dict = loads(metadata_str["geo"])
    geo_column = geo_dict["primary_column"]
    
    geometry_field_index = schema.get_field_index(geo_column)
    geometry_field = schema.field(geometry_field_index)
    geoarrow_geometry_field = geometry_field.with_metadata(
        {b"ARROW:extension:name": b"geoarrow.wkb"}
    )

    geoarrow_schema = schema.set(geometry_field_index, geoarrow_geometry_field)
    reader = RecordBatchReader.from_batches(geoarrow_schema, filtered_batches)
    gdf = GeoDataFrame.from_arrow(reader)
    gdf.set_crs("EPSG:4326",inplace=True)
    
    try:
        if columns:
            gdf = gdf[columns]
    except Exception as e:
        raise e
    
    return gdf

def read_parquet_arrow(path: str, region: str, 
                    columns: list[str] | None = None, 
                    filters: FilterStructure | None = None) -> GeoDataFrame:
    """
    Read tabular data from a parquet file on S3.
    
    Parameters:
    -----------
    path : str
        S3 path to the parquet file
    region : str
        AWS region
    columns : list, optional
        Columns to select
    filters : FilterStructure, optional
        Filter expression
        
    Returns:
    --------
    GeoDataFrame
        Filtered dataframe
    """
    
    # get pyarrow dataset
    clean_path = path.replace("s3://", "")
    try:
        ds = dataset(
            clean_path, filesystem=S3FileSystem(anonymous=True, region=region)
        )
    except Exception as e:
        raise S3ReadError(f"Read from bucket {clean_path} could not be complete.") from e
    
    try:
        batches = ds.to_batches()
    except Exception as e:
        exc_info = sys.exc_info()[0]
        catch_column_filter_error(exc_info,e)
        
    non_empty_batches = (b for b in batches if b.num_rows > 0)
    
    def filter_batches(batches):
        for b in batches:
            yield evaluate_filter_structure(b,filters)
    
    # generate results from complex filters(if needed)
    if filters:
        filtered_batches = filter_batches(non_empty_batches)
    else:
        filtered_batches = non_empty_batches
    
    schema = ds.schema
    reader = RecordBatchReader.from_batches(schema, filtered_batches)
    df = reader.read_pandas()
    if columns:
        df = df[columns]
    return df

def _get_gdf_from_bbox(release:str, bbox:tuple[float,float,float,float], columns:list[str], filters: FilterStructure, prefix: str, path: str, region: str):
    """Helper function to get a geodataframe from a bounding box."""
    main_path = path.format(release=release) + prefix
    gdf = read_geoparquet_arrow(main_path, region, bbox, columns=columns, filters=filters)
    return gdf

def from_address(address: str | tuple[float,float], prefix: str, main_path: str, region: str,
            release: str, columns: list[str]| None = None, filters: FilterStructure| None = None,
            distance: float = 500, unit: str = "m") -> GeoDataFrame:
    """
    Wrapper to geocode an address and fetch the geoparquet data within the address's area.
    
    Parameters:
    -----------
    address : str or tuple
        Address string or (longitude, latitude) tuple
    prefix : str
        Path prefix for the data
    main_path : str
        Base path template with {release} placeholder
    region : str
        AWS region
    release : str
        Release version
    columns : list, optional
        Columns to select
    filters : FilterStructure, optional
        Filter expression
    distance : float, default 500
        Buffer distance
    unit : str, default 'm'
        Unit of distance
        
    Returns:
    --------
    GeoDataFrame
        Filtered geodataframe
    """
    bbox = geocode_point_to_bbox(address, distance, unit)
    gdf = from_bbox(bbox,prefix,main_path,region,release,columns,filters)
    return gdf
    
def from_place(address: str, prefix: str, main_path: str, region: str, release: str,
            columns: list[str]| None=None, filters: FilterStructure| None=None) -> GeoDataFrame:
    """
    Wrapper to geocode a place and fetch the geoparquet data within the place.
    
    Parameters:
    -----------
    address : str
        Place name or address
    prefix : str
        Path prefix for the data
    main_path : str
        Base path template with {release} placeholder
    region : str
        AWS region
    release : str
        Release version
    columns : list, optional
        Columns to select
    filters : FilterStructure, optional
        Filter expression
        
    Returns:
    --------
    GeoDataFrame
        Filtered geodataframe
    """
    geometry, bbox = geocode_place_to_bbox(address)
    gdf = from_bbox(bbox,prefix,main_path,region,release,columns,filters)
    filtered_gdf = gdf[gdf.within(geometry)]
    return filtered_gdf

def from_bbox(bbox: tuple[float,float,float,float], prefix: str, main_path: str, region: str, 
            release: str, columns: list[str]| None=None, filters: FilterStructure| None=None) -> GeoDataFrame:
    """
    Wrapper to fetch the geoparquet data within the bounding box.
    
    Parameters:
    -----------
    bbox : tuple
        Bounding box as (minx, miny, maxx, maxy)
    prefix : str
        Path prefix for the data
    main_path : str
        Base path template with {release} placeholder
    region : str
        AWS region
    release : str
        Release version
    columns : list, optional
        Columns to select
    filters : FilterStructure, optional
        Filter expression
        
    Returns:
    --------
    GeoDataFrame
        Filtered geodataframe
    """
    gdf = _get_gdf_from_bbox(release, bbox, columns, filters, prefix, main_path, region)
    return gdf


from pyplaces import overture_maps as om, foursquare_open_places as fsq

# df=om.overture_base_from_place("Toronto","bathymetry")

# print(df)

print(fsq.find_categories("waffles"))

print(om.find_categories("waffles"))